#!/usr/bin/perl

$exec="/usr/sbin/kdb";
$cmdterminator="KTHREADS-TERMINATOR";
$verbose=1;
$kernel="";
$image="";
$sysstats=0;
$tracedata=0;
$outfile="";
@filters=();
$filterreg="";

$DEBUG=0;
$DEBUGARGS=0;
$DEBUGSEND=0;
$DEBUGWRITE=0;
$DEBUGPP=0;

@ESA=(
"exec_calls", 
"stub_calls", 
"gt1sec_calls", 
"gt10sec_calls", 
"gt60sec_calls", 
"longest_call", 
"longest_opnum", 
"longest_call_tstamp", 
"waycalls", 
"wayerrs", 
"gt10sec_waycalls",
"gt60sec_waycalls",
"gt120sec_waycalls",
"waycalls_queued",
"",
);
$ESS="10"; #hex

@PPA=(
"pkt_mutex", 
"version", 
"active_rqe", 
"active_xqe", 
"reservations", 
"mem_allocated", 
"mem_ration_level", 
"mem_peak_level", 
"max_reservations", 
"srv_reservations", 
"srv_resv_avail",
"noqueue_resv",
"active_ccall_resv",
"active_scall_resv",
"active_blocked_resv",
"total_reservations",
"total_sonly_resv",
"total_queued_resv",
"total_blocked_resv",
"total_noavail_resv",
"failed_alloc_rqe",
"blocked_alloc_rqe",
"rqe_overflows",
"xqe_mbuf_refs",
"active_rqes_peak",
"active_xqes_peak",
"xqe_throttles",
"mbuf_member_offset",
"fack_pkt_size",
"fack_pkt_size_increases",
"fack_pkt_size_decreases",
"large_pkt_size",
"larg_pkt_window_size",
"large_pkt_blast_size",
"low_pkts_window_size",
"pkt_gc_count",
"pkt_gc_timer",
"gc_pkts",
"gc_bytes",
"max_window",
"medium_window",
"small_window",
"low_pkts",
"rationing",
);

$PPS="28"; #this is hex

if( ProcessArgs(@ARGV) )
{
   usage;
   exit;
}

if( -e $outfile )
{
   print "Remove current $outfile (y/n)";
   $resp=<STDIN>;
   if( $resp =~ /^y/ )
   {
      unlink( $outfile );
      if( -e $outfile )
      {
         print "Unable to remove $outfile.\n";
         exit;
      }
   }
   else
   {
      print "Aborting...\n";
      exit;
   }
}

if( !open( OUTFILE, ">$outfile" ) )
{
   print "Can't open $outfile\n";
   exit;
}

if( $#filters < 0 )
{
   @filters=("."); #match everything
}

foreach (@filters)
{
   $filterreg .= "|" unless $filterreg eq "";
   $filterreg .= "($_)";
}

print "filterreg=[$filterreg]\n" if $DEBUG;

pipe(TODBREAD,TODBWRITE);

# create pipe from traffic FROM debugger

pipe(FROMDBREAD,FROMDBWRITE);

FORK:
if ($pid = fork) {
    &parent;
} elsif(defined $pid) {
    &child;
} elsif($1 =~ /No more process/) {
    sleep 5;
    redo FORK;
} else {
    die "$ME: can't fork\n";
}

# parent; write to TODBWRITE and read from FROMDBREAD

sub parent {

    close(TODBREAD);
    close(FROMDBWRITE);

    select(TODBWRITE); $| = 1;
    select(STDOUT); $| = 1;

    sendcmd("setup scroll");

    if( $tracedata )
    {
       qx(/opt/dcelocal/bin/dfstrace clear -log cmfx);
       qx(/opt/dcelocal/bin/dfstrace setset -set krpc -active);
    }
    if( $sysstats )
    {
       print "Getting data from kernel data structures...\n" if $verbose;
       &DumpPacketPool;
       &DumpExecStats;
       print "Gathered data from kernel data structures...\n" if $verbose;
    }
    if( $filterreg ne "(.)" )
    {
       print "Gathering filtered data about from system image.\n";
    }
    &FilterThreads;
    sendcmd("q");
    close(OUTFILE);
}

# child: read from TODBREAD and write to FROMDBWRITE

sub child {

    close(TODBWRITE);
    close(FROMDBREAD);

    close(STDIN) || die "$ME (child): Can't close STDIN\n";
    open(STDIN,"<&TODBREAD") || die "$ME (child): can't redirect STDIN\n";

    close(STDOUT);
    open(STDOUT,">&FROMDBWRITE") || die "$ME (child): can't redirect STDOUT\n";

    close(STDERR);
    open(STDERR,">&STDOUT")  || die "$ME (child): can't redirect STDERR\n";

    select(STDERR); $| = 1;
    select(STDOUT); $| = 1;

    exec "$exec $image $kernel";

    die "$ME: can't exec $exec\n";

}

sub sendcmd
{
   my ($cmd) =@_;
   my @out;
   print TODBWRITE "$cmd\n";
   print "sent: $cmd\n" if $DEBUGSEND;
   print TODBWRITE "!print $cmdterminator\n";
   print "sent: !print $cmdterminator\n" if $DEBUGSEND;
   @out=();
   while( <FROMDBREAD> )
   {
      print "got: $_" if $DEBUGSEND;
      next if /^\(\d+\)>/; #skip command string
      last if /^$cmdterminator/; #we're done
      @out=( @out, $_ );
   }
   return @out;
}

sub writeout
{
   my @lines = @_;

   foreach(@lines)
   {
      print OUTFILE "$_";
      print "$_" if $DEBUGWRITE;
   }
}

sub usage
{
   print <<EOF;
usage: kthreads [-k kernel] [-i system_image] [-q] [-s] [-t]
          -o output_file
     -k: specify the kernel image to use, default /unix
     -i: specify the system image to use, default current
     -q: quiet
     -s: gather system stats, specify the level of DCE
     -t: gather tracying output
EOF
}

#Do this manually instead of the nifty module cuz we don't
#know what perl modules will be on the box
sub ProcessArgs
{
   my (@args) = @_;
   my $arg;
   my $flag, $opt;

   while( $arg=shift @args )
   {
      print "arg: $arg\n" if $DEBUGARGS;
      $flag=$opt="";
      if( $arg !~ /^-(.)(.*)/ )
      {
         unshift @args, $arg;
         last;
      }
      $flag=$1;
      $opt=$2;
      print "flag=[$flag] opt=[$opt]\n" if $DEBUGARGS;
      return 1 if $flag eq "";
      $opt=shift @args if $opt eq "" && $flag =~/[kio]/;
      print "flag=[$flag] opt=[$opt]\n" if $DEBUGARGS;
      return 1 if $opt eq "" && $flag =~ /[kio]/;

      if( $flag eq "k" )
      {
         $kernel=$opt;  
         print "Using kernel image $kernel.\n" if $verbose;
         next;
      }
      if( $flag eq "i" )
      {
         $image=$opt;  
         print "Using system image $image.\n" if $verbose;
         next;
      }
      if( $flag eq "o" )
      {
         $outfile=$opt;  
         print "Using output file $outfile.\n" if $verbose;
         next;
      }
      if( $flag eq "q" )
      {
         $verbose=0;
         next;
      }
      if( $flag eq "s" )
      {
         $sysstats=1;  
         print "Going to gather system stats (netstat, iostat, vmstat).\n" 
            if $verbose;
         next;
      }
      if( $flag eq "t" )
      {
         $tracedata=1;
         print "Going to gather DFS tracing output.\n" if $verbose;
         next;
      }
   }
   if( $outfile eq "" )
   {
      print "Please specify an output file\n";
      return 1;
   }
   if( $kernel eq "" && $verbose )
   {
      print "Using the current system kernel.\n";
   }
   if( $image eq "" && $verbose )
   {
      print "Using the current system image.\n";
   }

   if( $#args > -1 )
   {
      @filters=@args;
      if( $verbose )
      {
         print "Using filters:";
         foreach (@filters)
         {
            print " $_";
         }
         print "\n";
      }
   }

   return 0;
}
         
sub FilterThreads
{
   my @thlist, @tmp, @out;
   my $line, $th;
   @thlist=sendcmd("thread *");
   @out=( "====  START: Thread information for system image  ====\n" );
   foreach $line ( @thlist )
   {
      print "line: $line" if $DEBUGFILTER;
      next unless $line =~ /$filterreg/;
      @tmp=split /\s+/, $line;
      $th=$tmp[1];
      $th =~ s/^(\d+).*/$1/;
      print "th: $th\n" if $DEBUGFILTER;    
      next unless $th =~ /^\d/;
      @out=( @out, "\n***** Thread info, slot $th *****\n" );
      @out=( @out, sendcmd("thread $th"), sendcmd("f $th") );
   }
   @out=( @out, "\n====   END: Thread information for system image   ====\n" );
   @out=( @out,   "======================================================\n" );
   writeout( @out );
}

sub DumpPacketPool
{
   my @lines;
   my @vals;
   my @cols;
   my @out;
   my $i;

   @out=();
   @out=( @out, "==============================================\n" );
   @out=( @out, "=======     Packet Pool Statistics     =======\n" );
   @out=( @out, "==============================================\n\n");
   writeout( @out );
   @out=();

   @lines=sendcmd("dw rpc_g_dg_pkt_pool $PPS");
   if( $lines[0] =~ /^[^\d]/ )
   {
      print "line: $line" if $DEBUGPP;
      writeout( "No match for rpc_g_dg_pkt_pool in dump\n" );
      return;
   }
   foreach (@lines)
   {
      @cols=split /\s+/;
      @vals=( @vals, hex($cols[1]), hex($cols[2]), 
              hex($cols[3]), hex($cols[4]) );
      print "$cols[1] $cols[2] $cols[3] $cols[4]\n" if $DEBUGPP;
   }
   for( $i=0; $i<=$#vals; $i+=2 )
   {
      @out=(@out, sprintf("%-24s:%-7d %-24s:%-7d\n",
                          $PPA[$i], $vals[$i],
                          $PPA[$i+1], $vals[$i+1]) );
   }
   writeout( @out, "\n" );
}

sub DumpExecStats
{
   my @lines;
   my @vals;
   my @cols;
   my @out;
   my $i;

   @out=();
   @out=( @out, "==============================================\n" );
   @out=( @out, "=======      Execution Statistics      =======\n" );
   @out=( @out, "==============================================\n\n");
   writeout( @out );
   @out=();

   @lines=sendcmd("dw rpc_g_dg_exec_stats $ESS");
   if( $lines[0] =~ /^[^\d]/ )
   {
      print "line: $line" if $DEBUGPP;
      writeout( "No match for rpc_g_dg_exec_stats in dump\n" );
      return;
   }
   foreach (@lines)
   {
      @cols=split /\s+/;
      @vals=( @vals, hex($cols[1]), hex($cols[2]), 
              hex($cols[3]), hex($cols[4]) );
      print "$cols[1] $cols[2] $cols[3] $cols[4]\n" if $DEBUGPP;
   }
   for( $i=0; $i<=$#ESA; $i+=2 )
   {
      @out=(@out, sprintf("%-24s:%-7d %-24s:%-7d\n",
                          $ESA[$i], $vals[$i],
                          $ESA[$i+1], $vals[$i+1]) );
   }
   writeout( @out, "\n" );
}

